/* 
   Linux loop encryption enabling module

   $Id: cryptoloop.c,v 1.2 2001/07/21 22:52:46 hvr Exp $

   Written by Herbert Valerio Riedel <hvr@gnu.org> 

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software

   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* use ./configure --enable-iv-mode-sector */
/* #define USE_LO_IV_MODE_SECTOR */

/* #define CRYPTOLOOP_DEBUG */

/* use atomic crypto functions */
/* #define CRYPTOLOOP_ATOMIC */

#include <linux/module.h>
#include <linux/version.h>

#ifdef CONFIG_KMOD
#include <linux/kmod.h>
#endif 

#include <linux/init.h>
#include <linux/config.h>
#include <linux/string.h>
#include <linux/crypto.h>
#include <linux/loop.h>
#include <linux/blk.h>

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,3)
# error you need at least kernel 2.4.3 -- unless you know exacty what you are doing
#endif

#if !defined(LO_CRYPT_CRYPTOAPI)
# warning LO_CRYPT_CRYPTOAPI not (yet) defined in kernel header
# define LO_CRYPT_CRYPTOAPI 18
#endif

static inline int
loop_get_bs(struct loop_device *lo)
{
  int bs = 0;

  if (blksize_size[MAJOR(lo->lo_device)])
    bs = blksize_size[MAJOR(lo->lo_device)][MINOR(lo->lo_device)];
  if (!bs)
    bs = BLOCK_SIZE;        

  return bs;
}

static int 
transfer_cryptoapi(struct loop_device *lo, int cmd, char *raw_buf,
                   char *loop_buf, int size, int real_block)
{ 
  struct cipher_context * cx = (struct cipher_context *) lo->key_data;
  struct cipher_implementation *ci = cx->ci;
  int (*encdecfunc)(struct cipher_context *cx, const u8 *in, u8 *out, 
		    int size, const u32 iv[]);
  char *in, *out;

  if (cmd == READ) {
#if defined(CRYPTOLOOP_ATOMIC)
    encdecfunc = ci->decrypt_atomic;
#else
    encdecfunc = ci->decrypt;
#endif
    in = raw_buf;
    out = loop_buf;
  } else {
#if defined(CRYPTOLOOP_ATOMIC)
    encdecfunc = ci->encrypt_atomic;
#else
    encdecfunc = ci->encrypt;
#endif
    in = loop_buf;
    out = raw_buf;
  }

#if defined(CRYPTOLOOP_DEBUG)
#warning CRYPTOLOOP_DEBUG enabled
  printk (KERN_DEBUG "cryptoxfer: cmd=%d size=%d iv=%d blksize=%d\n", 
	  cmd, size, real_block, loop_get_bs (lo));
#endif

#if defined(USE_LO_IV_MODE_SECTOR)
#warning USE_LO_IV_MODE_SECTOR enabled -- hope you know what this means
  /* split up transfer request into 512 byte data blocks */
  while (size > 0) {
    int _size = (size > LO_IV_SECTOR_SIZE) ? LO_IV_SECTOR_SIZE : size;    
    u32 iv[4] = { 0, };

    iv[0] = cpu_to_le32 (real_block);
    encdecfunc (cx, in, out, _size, iv);

    real_block++;
    size -= _size;
    in += _size;
    out += _size;
  }
#else
  {
    u32 iv[4] = { 0, };
    iv[0] = cpu_to_le32 (real_block);
    encdecfunc (cx, in, out, size, iv);
  }
#endif

  return 0;
}

static int 
cryptoapi_status(struct loop_device *lo, struct loop_info *info)
{
  int err = -EINVAL;
  char cipher[LO_NAME_SIZE];
  struct cipher_implementation *ci;
  struct cipher_context *cx;

#if defined(USE_LO_IV_MODE_SECTOR)
# if !defined(LO_IV_MODE_SECTOR)
# error You dont seem have a kernel patched for this to work...
# endif
  /* encryption breaks for non sector aligned offsets */
  if (info->lo_offset % LO_IV_SECTOR_SIZE) 
    goto out;

  lo->lo_iv_mode = LO_IV_MODE_SECTOR;
#endif

  strncpy(cipher, info->lo_name, LO_NAME_SIZE);
  cipher[LO_NAME_SIZE - 1] = 0;
                
  ci = find_cipher_by_name(cipher, 1);
  if (!ci)
    goto out;

  ci->lock();
  err = -ENOMEM;
  cx = ci->realloc_context(NULL, ci, info->lo_encrypt_key_size);
  if (!cx)
    goto out_ci;
  cx->ci = ci;
  ci->wipe_context(cx);
  if (ci->set_key(cx, info->lo_encrypt_key, 
		  info->lo_encrypt_key_size)) {
    err = -EINVAL;
    goto out_key;
  }
  lo->key_data = cx;
  err = 0;
  goto out;

 out_key:
  ci->wipe_context(cx);
  ci->free_context(cx);
 out_ci:
  ci->unlock();
 out:
  if (!err) {
    MOD_INC_USE_COUNT;
  }
	  
  return err;
}

static int 
cryptoapi_release(struct loop_device *lo)
{ 
  struct cipher_context *cx = (struct cipher_context *) lo->key_data;

  if (cx != NULL) { 
    cx->ci->wipe_context(cx);
    cx->ci->unlock();
    cx->ci->free_context(cx);
    lo->key_data=NULL;
    MOD_DEC_USE_COUNT;
    return 0;
  }

  printk (KERN_ERR "cryptoapi_release(): cx == NULL ????\n");
  return -EINVAL;
}

struct loop_func_table cryptoloop_funcs = { 
  number: LO_CRYPT_CRYPTOAPI,
  transfer: transfer_cryptoapi,
  init: cryptoapi_status,
  release: cryptoapi_release
};      

static int __init
init_cryptoloop(void)
{
  int rc;

  if ((rc = loop_register_transfer (&cryptoloop_funcs)))
    {
      printk(KERN_ERR "cryptoloop: register loop transfer function failed\n");
      return rc;
    }

  printk(KERN_INFO "cryptoloop: loaded\n");
  return 0;
}

static void __exit
cleanup_cryptoloop(void) 
{
  if (loop_unregister_transfer (LO_CRYPT_CRYPTOAPI))
    printk(KERN_ERR "cryptoloop: unregistering transfer funcs failed\n");

  printk(KERN_INFO "cryptoloop: unloaded\n");
}

module_init(init_cryptoloop);
module_exit(cleanup_cryptoloop);

/* 
 * Local variables:
 *  c-file-style: "gnu"
 *  tab-width: 8
 *  indent-tabs-mode: nil
 * End:
 */
